#!/usr/bin/env python

'''Super-Simple Scheduler for BG/L'''
__revision__ = '$Revision'

from sss.restriction import Data, DataSet
from sss.server import EventReceiver, Server
from elementtree.ElementTree import XML, Element
from syslog import syslog, LOG_INFO, LOG_ERR
from time import time
import sys

sys.path.append('/soft/apps/rm-0.90/lib/python')
import DB2

class DataSet2(DataSet):
    '''DataSet2 is DataSet with class parameters'''
    rname = None
    ename = None
    obj = None
    idalloc = None
    log = False
    
    def __init__(self):
        self.data = []

class FailureMode(object):
    '''FailureModes are used to report (and supress) errors appropriately
    call Pass() on success and Fail() on error'''
    def __init__(self, name):
        self.name = name
        self.status = True

    def Pass(self):
        '''Check if status was previously failure and report OK status if needed'''
        if not self.status:
            syslog(LOG_ERR, "Failure %s cleared" % (self.name))
            self.status = True

    def Fail(self):
        '''Check if status was previously success and report failed status if needed'''
        if self.status:
            syslog(LOG_ERR, "Failure %s occured" % (self.name))
            self.status = False

class Partition(Data):
    '''Partitions are allocatable chunks of the machine'''
    def __init__(self, element):
        Data.__init__(self, element)
        self.element.set('state', 'idle')
        self.job = 'none'
        self.rcounter = 1

    def CheckDBState(self):
        #syslog(LOG_INFO, "Starting DB check for partition %s" % (self.element.get('name')))
        conn = DB2.connect(uid='bglsysdb', pwd='db24bgls', dsn='bgdb0')
        cursor = conn.cursor()
        cursor.execute("select status from bglblock where blockid='%s' or blockid='%s_vn';" % ((self.element.get('name'),) *2))
        result = cursor.fetchall()
        #syslog(LOG_INFO, "Query finished")
        if result.count(('F',)) == len(result):
            cursor.execute('''update bglblock set MODE='V'where blockid='%s_vn';''' % (self.element.get('name')))
            conn.close()
            return True
        else:
            conn.close()
            syslog(LOG_ERR, "Found confused partition %s" % (self.element.get('name')))
            return False

    def CanRun(self, job):
        '''Check that job can run on partition with reservation constraints'''
        # check admin
        if self.element.get('admin', None) != 'online':
            return False
        # check size
        jobsize = int(job.element.get('nodes'))
        if self.element.get('size') == '32':
            if not 0 < jobsize <= 32:
                return False
        elif self.element.get('size') == '64':
            if not 32 < jobsize <= 64:
                return False
        elif self.element.get('size') == '128':
            if not 64 < jobsize <= 128:
                return False
        elif self.element.get('size') == '256':
            if not 128 < jobsize <= 256:
                return False
        elif self.element.get('size') == '512':
            if not 256 < jobsize <= 512:
                return False
        elif self.element.get('size') == '1024':
            if not 512 < jobsize <= 1024:
                return False
        else:
            syslog(LOG_INFO, "Unsupported partition size %s for partition %s" % (self.element.get('size'),
                                                                                 self.element.get('name')))
            return False

        if job.element.get('queue') not in self.element.get('queue').split(':') + ['BUG']:
            return False
        # add a little slack for job cleanup with reservation calculations
        wall = float(job.element.get('walltime')) + 5.0
        jdur = 60 * wall
        # all times are in seconds
        current = time()
        for reserv in self.element.findall('Reservation'):
            if job.element.get('user') in reserv.get('user', '').split(':'):
                continue
            start = float(reserv.get('start'))
            rdur = int(reserv.get('duration'))
            if current < start:
                # reservation has not started
                if start < (current + jdur):
                    return False
            elif current > (start + rdur):
                # reservation has finished
                continue
            else:
                # reservation is active
                return False
        if not self.CheckDBState():
            return False
        return True

    def PlaceJob(self, job):
        '''Allocate this partition for Job'''
        syslog(LOG_INFO, "Scheduling job %s on partition %s" % (job.element.get('jobid'),
                                                                self.element.get('name')))
        self.job = job.element.get('jobid')
        self.element.set('state', 'busy')

    def Free(self):
        '''DeAllocate partition for current job'''
        syslog(LOG_INFO, "Freeing partition %s" % (self.element.get('name')))
        self.job = 'none'
        self.element.set('state', 'idle')

    def AddReservation(self, args):
        '''Add a reservation for this partition'''
        reservation = Element('Reservation')
        reservation.attrib.update(args)
        if not args.has_key('name'):
            reservation.set('name', "%s.%s" % (self.element.get('name'), self.rcounter))
            self.rcounter += 1
        self.element.append(reservation)

    def DelReservation(self, name):
        [self.element.remove(reservation) for reservation in self.element.findall('Reservation') if reservation.get('name') == name]

class PartitionSet(DataSet2):
    rname = 'Partitions'
    ename = 'Partition'
    obj = Partition
        
class Job(Data):
    '''This class is represents User Jobs'''
    def __init__(self, element):
        Data.__init__(self, element)
        self.partition = 'none'
        syslog(LOG_INFO, "Found new job %s" % (self.element.get('jobid')))

    def Place(self, partition):
        '''Build linkage to execution partition'''
        self.partition = partition.element.get('name')
        self.element.set('state', 'running')

class BGSched(Server):
    '''This scheduler implements a fifo policy'''
    subscriptions = [('queue-manager', '**', '**')]
    __implementation__ = 'BGSched'
    __component__ = 'scheduler'
    __dispatch__ = {'events':'HandleEvent',
                    'GetPartition':'partitions.Get', 'AddPartition':'partitions.Add',
                    'DelPartition':'partitions.Del', 'AddReservation':'AddRes',
                    'DelReservation':'DelRes', 'Online':'Online', 'Offline':'Offline',
                    'SetQueue': 'SetQueue'}
    __statefields__ = ['partitions']
    __schedcycle__ = 60

    def AddRes(self, xml, portinfo):
        '''Handler for addition of reservations'''
        return self.partitions.Get(xml, portinfo, self.AddRes_cb)

    def AddRes_cb(self, partition, args):
        '''Callback for reservation addition'''
        partition.AddReservation(args)
    
    def DelRes(self, xml, portinfo):
        '''Handler for reservation deletion'''
        return self.partitions.Get(xml, portinfo, self.DelRes_cb)

    def DelRes_cb(self, partition, args):
        '''Callback for reservation deletion'''
        partition.DelReservation(args['name'])

    def Online(self, xml, portinfo):
        '''Handler for partition online'''
        return self.partitions.Get(xml, portinfo, self.Online_cb)

    def Online_cb(self, partition, args):
        '''Callback for partition online'''
        partition.element.set('admin', 'online')

    def Offline(self, xml, portinfo):
        '''Handler for partition offline'''
        return self.partitions.Get(xml, portinfo, self.Offline_cb)

    def Offline_cb(self, partition, args):
        '''Callback for partition online'''
        partition.element.set('admin', 'offline')

    def SetQueue(self, xml, portinfo):
        '''Handler for setting queue list for partitions'''
        return self.partitions.Get(xml, portinfo, self.SetQueue_cb)

    def SetQueue_cb(self, partition, args):
        '''Callback for queue setting'''
        partition.element.set('queue', args['queue'])

    def __setup__(self):
        self.partitions = PartitionSet()
        self.jobs = []
        self.qmconnect = FailureMode("QM Connection")
        self.lastrun = 0

    def __progress__(self):
        since = time() - self.lastrun
        if since > self.__schedcycle__:
            self.RunQueue()
            self.lastrun = time()
        return 0
            
    def RunQueue(self):
        try:
            handle = self.comm.ClientInit('queue-manager')
        except:
            self.qmconnect.Fail()
            return 0
        self.qmconnect.Pass()
        self.comm.SendMessage(handle,
                              "<get-job><job nodes='*' user='*' jobid='*' state='*' walltime='*' queue='*'/></get-job>")
        jobs = self.comm.RecvMessage(handle)
        xjobs = XML(jobs)
        active = [job.get('jobid') for job in xjobs.findall(".//job")]
        for job in [j for j in self.jobs if j.element.get('jobid') not in active]:
            syslog(LOG_INFO, "Job %s gone from qm" % (job.element.get('jobid')))
            self.jobs.remove(job)
        # known is jobs that are already registered
        known = [job.element.get('jobid') for job in self.jobs]
        [partition.Free() for partition in self.partitions if partition.job not in known + ['none']]
        newjobs = [job for job in xjobs.findall('.//job') if job.get('jobid') not in known]
        self.jobs.extend([Job(job) for job in newjobs])
        # for free partitions
        for idle in [part for part in self.partitions if part.element.get('state') == 'idle']:
            # with jobs that can run
            for job in [job for job in self.jobs if ((job.element.get('state') == 'queued') and idle.CanRun(job))]:
                idle.PlaceJob(job)
                job.Place(idle)
                self.comm.SendMessage(handle,
                                      "<run-job nodelist='%s'><job jobid='%s'/></run-job>" %
                                      (idle.element.get('name'), job.element.get('jobid')))
                self.comm.RecvMessage(handle)
                break
        self.comm.ClientClose(handle)
        return 0
                
if __name__ == '__main__':
    from getopt import getopt, GetoptError
    try:
        (opts, arguments) = getopt(sys.argv[1:], 'd', ['daemon='])
    except GetoptError, msg:
        print "%s\nUsage:\nbgsched.py [-d] [--daemon <pidfile>]" % (msg)
        raise SystemExit, 1
    daemon = [x[1] for x in opts if x[0] == '--daemon']
    debug = len([x for x in opts if x[0] == '-d'])
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0])
    server = BGSched(debug=debug)
    server.ServeForever()
    

